# -*- coding: utf-8 -*-
# :Project:   hurm -- Subcountries data view
# :Created:   mar 02 feb 2016 19:47:59 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from operator import itemgetter

from pyramid.view import view_config


@view_config(route_name="subcountries", renderer="json")
def _subcountries(request):
    from gettext import translation
    from pycountry import LOCALES_DIR, subdivisions

    country = request.params.get('country', 'IT')

    lname = request.locale_name
    try:
        t = translation('iso3166_2', LOCALES_DIR, languages=[lname]).gettext
    except IOError:
        t = lambda x: x

    result = [dict(code=s.code, name=t(s.name))
              for s in subdivisions.get(country_code=country)
              if country != 'IT' or s.type == 'Province']

    result.sort(key=itemgetter('name'))

    return dict(count=len(result),
                message="Ok",
                root=result,
                success=True)



# -*- coding: utf-8 -*-
# :Project:   hurm -- Location related printouts
# :Created:   sab 13 feb 2016 19:48:24 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from itertools import groupby

from pyramid.view import view_config

from reportlab.lib.pagesizes import A4, landscape
from reportlab.platypus import TableStyle
from reportlab.platypus.tables import Table

from sqlalchemy.orm import object_session

from hurm.db.entities import Duty, Location, Task

from .. import DBSession
from ..i18n import translatable_string as _
from . import BasicEditionPrintout, ParameterError, create_pdf
from .calendar import CellContent, calendar_header, calendar_row


def _duties_to_cells(translate, duties):
    duties = list(duties)

    timings = []
    for duty in duties:
        t = duty.starttime
        if t not in timings:
            timings.append(t)
        t = duty.endtime
        if t not in timings:
            timings.append(t)

    timings.sort()

    cells = []
    for i, t in enumerate(timings[:-1]):
        cells.append(CellContent(t, timings[i+1], []))

    for cell in cells:
        cduties = [duty for duty in duties
                   if ((cell.starttime <= duty.starttime < cell.endtime)
                       or
                       (duty.starttime < cell.endtime <= duty.endtime))]
        if cell.endtime.hour - cell.starttime.hour > 1:
            persons = (translate(duty.person.fullname) for duty in cduties)
        else:
            persons = (translate(duty.person.abbreviated_fullname) for duty in cduties)
        cell.content.extend(sorted(persons))

    cells = [cell for cell in cells if cell.content]

    return cells


def duties_table(translate, document, location, edition):
    # TRANSLATORS: this is a Python strftime() format, see
    # http://docs.python.org/3/library/time.html#time.strftime
    datefmt = translate(_('%m-%d-%Y'))

    # TRANSLATORS: this is the Python strftime() template used to format times,
    # see http://docs.python.org/3/library/time.html#time.strftime
    timefmt = translate(_('%I:%M %p'))

    # TRANSLATORS: this is the strftime() template used to format hours,
    # see http://docs.python.org/3/library/time.html#time.strftime
    hourfmt = translate(_('%I %p'))

    session = object_session(location)
    allduties = session.query(Duty).join(Task) \
                .filter(Task.idlocation == location.idlocation) \
                .filter(Task.idedition == edition.idedition) \
                .order_by(Task.date, Duty.starttime) \
                .all()
    if allduties:
        min_time = min(duty.starttime for duty in allduties)
        if min_time.minute != 0:
            min_time = min_time.replace(minute=0)
        max_time = max(duty.endtime for duty in allduties)
        if max_time.minute != 0:
            max_time = max_time.replace(hour=max_time.hour+1, minute=0)

        cells, styles = calendar_header(hourfmt, font_size=4,
                                        min_time=min_time, max_time=max_time)
    else:
        cells, styles = calendar_header(hourfmt, font_size=4)

    rows = [cells]
    rownumber = 1

    task_date = lambda d: d.task.date

    for date, duties in groupby(allduties, task_date):
        ccells = _duties_to_cells(translate, duties)
        for cell in ccells:
            if (cell.endtime.hour - cell.starttime.hour) > 1:
                cell.content.insert(0,
                                    cell.starttime.strftime(timefmt)
                                    + " — "
                                    + cell.endtime.strftime(timefmt))
            else:
                cell.content.insert(0, cell.starttime.strftime(timefmt))
                cell.content.insert(1, cell.endtime.strftime(timefmt))
        cells, rowstyles = calendar_row(rownumber, ccells,
                                        min_time=min_time, max_time=max_time)
        rows.append([date.strftime(datefmt)] + cells)
        styles.extend(rowstyles)
        rownumber += 1

    ntimecells = len(rows[0]) - 1
    cwidths = [None] + [(document.width - 40) // ntimecells] * ntimecells

    return Table(rows, cwidths, style=TableStyle(styles), repeatRows=1)


class LocationDuties(BasicEditionPrintout):
    pagesize = landscape(A4)

    @classmethod
    def getArgumentsFromRequest(klass, session, request):
        args = super().getArgumentsFromRequest(session, request)

        t = args[0]

        try:
            idlocation = int(request.matchdict['idlocation'])
        except KeyError:
            raise ParameterError(t(_('Missing argument: $name',
                                     mapping=dict(name='idlocation'))))
        location = session.query(Location).get(idlocation)
        if location is None:
            raise ParameterError(t(_('No location with id $id',
                                 mapping=dict(id=str(idlocation)))))

        args.append(location)

        return args

    def __init__(self, output, translate, pluralize, edition, location):
        super().__init__(output, translate, pluralize, edition)
        self.location = location

    def getCenterHeader(self):
        return self.translate(_('Duties calendar for $location',
                                mapping=dict(location=self.location.description)))

    def getTitle(self):
        return self.location.description

    def getSubTitle(self):
        session = object_session(self.location)
        count = session.query(Duty).join(Task) \
                .filter(Task.idedition == self.edition.idedition) \
                .count()
        ngettext = self.pluralize
        return ngettext('$count duty', '$count duties', count,
                        mapping=dict(count=count))

    def getElements(self):
        yield from super().getElements()
        yield duties_table(self.translate, self.doc, self.location, self.edition)


@view_config(route_name='pdf_location_duties')
def _location_duties(request):
    return create_pdf(DBSession(), request, LocationDuties)



# -*- coding: utf-8 -*-
# :Project:   hurm -- Functional tests configuration
# :Created:   dom 07 feb 2016 18:23:13 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from pyramid.paster import get_appsettings
from webtest import TestApp
import pytest

from hurm.db import entities
from hurm.fe import main


@pytest.fixture(scope="module")
def app():
    settings = get_appsettings('test.ini')
    app = TestApp(main({}, **settings))
    app.post('/auth/login', {'email': 'bob@example.com', 'password': 'test'})
    return app


@pytest.fixture(scope="module")
def session():
    return entities.DBSession()


@pytest.fixture
def test_edition(session):
    return session.query(entities.Edition).filter_by(description='Test edition').one()


@pytest.fixture
def jane_tree(session):
    return session.query(entities.Person).filter_by(lastname='Tree').one()


@pytest.fixture
def reception(session):
    return session.query(entities.Location).filter_by(description='Reception').one()



# -*- coding: utf-8 -*-
# :Project:   hurm -- Persons data view
# :Created:   mar 02 feb 2016 09:58:18 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from pyramid.view import view_config

from sqlalchemy import bindparam, cast, desc, exists, func, select, text, Text
from sqlalchemy.sql.expression import literal_column
from sqlalchemy.types import String

from hurm.db.tables import (
    activities,
    availabilities,
    duties,
    persons,
    personactivities,
    tasks
    )

from ..i18n import translatable_string as _
from . import expose


_a = availabilities.alias('a')
_d = duties.alias('d')
_p = persons.alias('p')
_t = tasks.alias('t')
_pa = personactivities.alias('pa')
_act = activities.alias('act')

@view_config(route_name='persons', renderer='json')
@expose(select([_p.c.idperson,
                _p.c.lastname,
                _p.c.firstname,
                # text() is needed here, to workaround default normalization
                # performed by the Name type decorator, sigh...
                (_p.c.lastname + text("' '") + _p.c.firstname).label('FullName'),
                _p.c.phone,
                _p.c.mobile,
                _p.c.email,
                _p.c.birthdate,
                _p.c.role,
                func.string_agg(cast(_act.c.idactivity, Text), ',').label('idactivities'),
                func.string_agg(_act.c.description, ', ').label('Activities'),
                _p.c.note,
                literal_column("'*'", String).label('password'),
                select([func.count(_d.c.idduty)],
                       _d.c.idperson == _p.c.idperson)
                .as_scalar().label('Duties'),
                select([func.count(_a.c.idavailability)],
                       _a.c.idperson == _p.c.idperson)
                .as_scalar().label('Availabilities')])
        .select_from(_p.outerjoin(_pa.join(_act)))
        .group_by(_p.c.idperson),
        metadata=dict(
            Activities=dict(
                label=_('Activities'),
                hint=_('Preferred activities of the person.'),
                width=140,
            ),
            Availabilities=dict(
                hidden=True,
                label=_('Availabilities'),
                hint=_('Number of availabilities of the person, in any edition.')
            ),
            Duties=dict(
                hidden=True,
                label=_('Duties'),
                hint=_('Number of duties assigned to the person, in any edition.')
            ),
            FullName=dict(
                flex=1,
                hint=_('Full name of the person'),
                label=_('Full name'),
            ),
            birthdate=dict(hidden=True),
            email=dict(width=140, vtype='email'),
            idactivities=dict(
                hidden=True,
                nullable=True,
                label=_('Activities'),
                hint=_('Preferred activities of the person.'),
            ),
            firstname=dict(hidden=True),
            lastname=dict(hidden=True),
            mobile=dict(hidden=True, vtype='phone'),
            note=dict(width=120),
            password=dict(
                hidden=True,
                hint=_('Login password of the user'),
                label=_('Password'),
                nullable=True,
                password=True,
                width=70,
            ),
            phone=dict(hidden=True, vtype='phone'),
            role=dict(hidden=True, width=140),
        ))
def _persons(request, results):
    return results


_t2 = tasks.alias('t2')

@view_config(route_name='available_persons', renderer='json')
@expose(select([_p.c.idperson,
                # text() is needed here, to workaround default normalization
                # performed by the Name type decorator, sigh...
                (_p.c.lastname + text("' '") + _p.c.firstname).label('fullname'),
                exists()
                .where(_pa.c.idperson == _p.c.idperson)
                .where(_t2.c.idtask == bindparam('idtask'))
                .where(_t2.c.idactivity == _pa.c.idactivity)
                .label('Preferred')])
        .order_by(desc('Preferred'), 'lastname', 'firstname'))
def _available_persons():
    request, params = (yield)
    if 'idtask' in params:
        idtask = int(params.pop('idtask'))
        bparams = params.setdefault('params', {})
        bparams['idtask'] = idtask
        conditions = (exists()
                      .where(_t.c.idtask == idtask)
                      .where(_a.c.date == _t.c.date)
                      .where(_a.c.idperson == _p.c.idperson)
                      .where(text("(COALESCE(a.starttime, '00:00'),"
                                  " COALESCE(a.endtime, '24:00'))"
                                  " OVERLAPS "
                                  "(t.starttime,"
                                  " COALESCE(t.endtime, '24:00'))")),)
        result = yield params, conditions
    else:
        bparams = params.setdefault('params', {})
        bparams['idtask'] = None
        result = yield params
    yield result



# -*- coding: utf-8 -*-
# :Project:   hurm -- Pyramid entry point
# :Created:   lun 01 feb 2016 20:21:56 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

import logging

from pyramid.config import Configurator
from pyramid.session import SignedCookieSessionFactory

from sqlalchemy import engine_from_config, exc

import zope.sqlalchemy

from hurm.db.entities import DBSession

logger = logging.getLogger(__name__)


def exception_view(context, request):
    from nssjson import dumps
    from pyramid.response import Response

    logger.error('Error happened serving %s: %s, %s',
                 request.url, context.status, context)
    return Response(dumps({'success': False, 'message': str(context)}),
                    status=context.status, content_type='application/json')


def db_exception_tween_factory(handler, registry):
    from nssjson import dumps
    from pyramid.response import Response

    def tween(request):
        try:
            response = handler(request)
        except exc.SQLAlchemyError as e:
            logger.error('Error happened serving %s: %s', request.url, e)
            return Response(dumps({'success': False, 'message': str(e)}),
                            # status=500,
                            content_type='application/json')
        else:
            return response

    return tween


def main(global_config, **settings):
    """ This function returns a Pyramid WSGI application.
    """

    from metapensiero.extjs.desktop.pyramid import configure

    engine = engine_from_config(settings, 'sqlalchemy.')
    DBSession.configure(bind=engine)

    zope.sqlalchemy.register(DBSession)

    timeout = settings.get('session.timeout', 24*60*60)
    if timeout == 'None':
        timeout = None
    else:
        timeout = int(timeout)

    reissue_time = settings.get('session.reissue_time', 24*60*60)
    if reissue_time == 'None':
        reissue_time = None
    else:
        reissue_time = int(reissue_time)

    session_factory = SignedCookieSessionFactory(settings['session.secret'],
                                                 timeout=timeout,
                                                 reissue_time=reissue_time)
    config = Configurator(settings=settings,
                          session_factory=session_factory,
                          exceptionresponse_view=exception_view)
    config.add_tween('hurm.fe.db_exception_tween_factory')

    configure(config)

    config.add_translation_dirs('hurm.db:locale/')
    config.add_translation_dirs('hurm.fe:locale/')

    config.add_static_view('static', 'static', cache_max_age=60*60)

    config.add_route('login', '/auth/login')
    config.add_route('logout', '/auth/logout')

    config.add_route('activities', '/data/activities')
    config.add_route('availabilities', '/data/availabilities')
    config.add_route('available_persons', '/data/available_persons')
    config.add_route('countries', '/data/countries')
    config.add_route('duties', '/data/duties')
    config.add_route('editions', '/data/editions')
    config.add_route('locations', '/data/locations')
    config.add_route('persons', '/data/persons')
    config.add_route('subcountries', '/data/subcountries')
    config.add_route('tasks', '/data/tasks')

    config.add_route('pdf_by_location_duties',
                     '/pdf/duties/edition/{idedition}/locations')
    config.add_route('pdf_by_person_duties',
                     '/pdf/duties/edition/{idedition}/persons')
    config.add_route('pdf_location_duties',
                     '/pdf/duties/edition/{idedition}/location/{idlocation}')
    config.add_route('pdf_person_duties',
                     '/pdf/duties/edition/{idedition}/person/{idperson}')
    config.add_route('pdf_persons_duties_summary',
                     '/pdf/duties/edition/{idedition}/summary')

    config.add_route('yaml_dump', '/yaml/edition/{idedition}')
    config.add_route('excel_dump', '/xls/edition/{idedition}')

    config.scan()

    return config.make_wsgi_app()



# -*- coding: utf-8 -*-
# :Project:   hurm -- Countries data view
# :Created:   mar 02 feb 2016 19:05:05 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from operator import itemgetter

from pyramid.view import view_config


@view_config(route_name="countries", renderer="json")
def _countries(request):
    from gettext import translation
    from pycountry import LOCALES_DIR, countries

    lname = request.locale_name
    try:
        t = translation('iso3166', LOCALES_DIR, languages=[lname]).gettext
    except IOError:
        t = lambda x: x
    result = [dict(code=c.alpha2, name=t(c.name)) for c in countries]

    result.sort(key=itemgetter('name'))

    return dict(count=len(result),
                message="Ok",
                root=result,
                success=True)



# -*- coding: utf-8 -*-
# :Project:   hurm -- Views declaration
# :Created:   lun 01 feb 2016 20:54:33 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

import logging

from metapensiero.sqlalchemy.proxy.pyramid import expose

from hurm.db.entities import DBSession
from hurm.db.utils import save_changes


logger = logging.getLogger(__name__)
changes_logger = logging.getLogger(__name__ + '-changes')


class LoggerAdapter(logging.LoggerAdapter):
    "Add username and remote IP to the logged message"

    def process(self, msg, kwargs):
        extra = self.extra
        msg = "[%s@%s] " % (extra['user'], extra['ip']) + msg
        return msg, kwargs


def get_request_logger(request, logger):
    "Get a specialized `logger` for a Pyramid `request`"

    extra = dict(
        ip=request.client_addr,
        user=request.session.get('user_fullname') or 'anonymous')

    return LoggerAdapter(logger, extra)


def logged_save_changes(sasess, request, modified, deleted):
    "Save changes logging changes with user information"

    return save_changes(sasess, request, modified, deleted,
                        clogger=get_request_logger(request, changes_logger))


# Configure the `expose` decorator
expose.create_session = staticmethod(lambda request: DBSession())
expose.save_changes = staticmethod(logged_save_changes)



# -*- coding: utf-8 -*-
# :Project:   hurm -- Abstract printouts
# :Created:   gio 11 feb 2016 17:14:12 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from copy import copy
from datetime import datetime
import logging
from os import unlink
from tempfile import mktemp

from pyramid.httpexceptions import HTTPBadRequest, HTTPInternalServerError


logger = logging.getLogger(__name__)

BASE_FONT_NAME = 'DejaVuSans'

from reportlab import rl_settings
rl_settings.canvas_basefontname = BASE_FONT_NAME

from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont, TTFError

try:
    for variant in ('', '-Bold', '-Oblique', '-BoldOblique'):
        pdfmetrics.registerFont(TTFont(BASE_FONT_NAME + variant,
                                       BASE_FONT_NAME + "%s.ttf" % variant))
except TTFError: # pragma: no cover
    from reportlab import rl_config

    logger.error('Could not find the "%s" font, using PDF default fonts', BASE_FONT_NAME)

    BASE_FONT_NAME = 'Times-Roman'
    rl_config.canvas_basefontname = rl_settings.canvas_basefontname = BASE_FONT_NAME
    BOLD_ITALIC_FONT_NAME = 'Times-BoldItalic'
    ITALIC_FONT_NAME = 'Times-Italic'
else:
    BOLD_ITALIC_FONT_NAME = BASE_FONT_NAME + '-BoldOblique'
    ITALIC_FONT_NAME = BASE_FONT_NAME + '-Oblique'

    from reportlab.lib.fonts import addMapping
    addMapping(BASE_FONT_NAME, 0, 0, BASE_FONT_NAME)
    addMapping(BASE_FONT_NAME, 0, 1, BASE_FONT_NAME + '-Oblique')
    addMapping(BASE_FONT_NAME, 1, 0, BASE_FONT_NAME + '-Bold')
    addMapping(BASE_FONT_NAME, 1, 1, BASE_FONT_NAME + '-BoldOblique')

from reportlab.lib.enums import TA_CENTER
from reportlab.lib.pagesizes import A4
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.lib.units import cm
from reportlab.platypus import (
    BaseDocTemplate,
    Frame,
    FrameBreak,
    NextPageTemplate,
    PageTemplate,
    Paragraph,
    )

from hurm.db.entities import Edition

from ..i18n import gettext, pluralizer, translatable_string as _, translator



base_style = getSampleStyleSheet()
"The base style used to build the document"

title_style = copy(base_style['Title'])
"The style used for the title of the document"

title_style.fontSize = 28
title_style.leading = title_style.fontSize*1.1

subtitle_style = copy(base_style['Heading1'])
"The style used for the subtitle of the document"

subtitle_style.fontSize = 20
subtitle_style.leading = subtitle_style.fontSize*1.1
subtitle_style.alignment = TA_CENTER
subtitle_style.fontName = ITALIC_FONT_NAME

heading_style = copy(base_style['Heading2'])
"The style used for the heading paragraphs of the document"

heading_style.alignment = TA_CENTER

normal_style = copy(base_style['Normal'])
"The style used for most of the paragraphs of the document"

normal_style.fontSize = 14
normal_style.leading = normal_style.fontSize*1.1

caption_style = copy(base_style['Italic'])
"The style used for the caption of the table's columns"

caption_style.fontSize = 9
caption_style.leading = caption_style.fontSize*1.1


def reduce_fontsize_to_fit_width(text, maxwidth, *styles):
    """Reduce the font size of the given styles to fit a max width.

    :param text: the string of text
    :param maxwidth: maximum width that can be used
    :param styles: the list of styles that should be adapted
    :returns: a list of (copies of) the styles with the adapted font size
    """

    from reportlab.pdfbase.pdfmetrics import stringWidth

    copies = styles
    mainstyle = styles[0]

    while stringWidth(text, mainstyle.fontName, mainstyle.fontSize) > maxwidth:
        if mainstyle is styles[0]: # pragma: no cover
            copies = [copy(style) for style in styles]
            mainstyle = copies[0]

        for style in copies:
            style.fontSize -= 1
            style.leading = style.fontSize * 1.1

    return copies


class BasicPrintout(object):
    """Abstract base class used to implement the printouts.

    This class implements the logic used by most printouts, producing a PDF document in the
    `output` filename.

    The document has a front page with an header, a body splitted into `columns` frames and
    a footer. Succeding pages do not have the header frame.
    """

    leftMargin = 1*cm
    "The width of the left margin, by default 1cm"

    rightMargin = 1*cm
    "The width of the right margin, by default 1cm"

    topMargin = 1*cm
    "The width of the top margin, by default 1cm"

    bottomMargin = 1*cm
    "The width of the bottom margin, by default 1cm"

    pagesize = A4
    "The page size, by default A4 in portrait orientation"

    @classmethod
    def getArgumentsFromRequest(klass, session, request):
        """Extract needed arguments for the constructor from the request.

        :param session: the SQLAlchemy session
        :param request: the Pyramid request instance
        :rtype: a sequence of arguments, the first one being a ``translate`` function
        """

        return [translator(request), pluralizer(request)]

    def __init__(self, output, translate, pluralize, columns=1):
        """Initialize the instance.

        :param output: a filename where the PDF will be written
        :param columns: number of columns
        """

        self.output = output
        self.translate = translate
        self.pluralize = pluralize
        self.columns = columns
        self.timestamp = datetime.now()

    @property
    def cache_max_age(self):
        "Compute the cache control max age, in seconds."

        return 0

    def createDocument(self):
        """Create the base Platypus document."""

        from pkg_resources import get_distribution

        version = get_distribution('hurm.fe').version

        doc = self.doc = BaseDocTemplate(
            self.output, pagesize=self.pagesize, showBoundary=0,
            leftMargin=self.leftMargin, rightMargin=self.rightMargin,
            topMargin=self.topMargin, bottomMargin=self.bottomMargin,
            author='HuRM %s' % version,
            creator="https://bitbucket.org/lele/hurm",
            subject=self.__class__.__name__,
            title='%s: %s' % (self.getTitle(), self.getSubTitle()))

        title_height = 3.0*cm
        title_width = doc.width
        title_frame = Frame(doc.leftMargin, doc.height + doc.bottomMargin - title_height,
                            title_width, title_height)
        self.title_width = title_width

        fp_frames = [title_frame]
        lp_frames = []

        fwidth = doc.width / self.columns
        fheight = doc.height

        bmargin = doc.bottomMargin
        for f in range(self.columns):
            lmargin = doc.leftMargin + f*fwidth
            fp_frames.append(Frame(lmargin, bmargin, fwidth, fheight-title_height,
                                   showBoundary=0))
            lp_frames.append(Frame(lmargin, bmargin, fwidth, fheight, showBoundary=0))

        templates = [PageTemplate(frames=fp_frames,
                                  id="firstPage",
                                  onPage=self.decoratePage),
                     PageTemplate(frames=lp_frames,
                                  id="laterPages",
                                  onPage=self.decoratePage)]
        doc.addPageTemplates(templates)

    def getLeftHeader(self):
        "The top left text."

        raise NotImplementedError("%s should implement this method!"
                                  % self.__class__)

    def getRightHeader(self):
        "The top right text."

        raise NotImplementedError("%s should implement this method!"
                                  % self.__class__)

    def getCenterHeader(self):
        "The top center text."
        # pragma: no cover
        raise NotImplementedError("%s should implement this method!"
                                  % self.__class__)

    def getTitle(self):
        "The title of the document."

        raise NotImplementedError("%s should implement this method!"
                                  % self.__class__)

    def getSubTitle(self):
        "The subtitle of the document."

        raise NotImplementedError("%s should implement this method!"
                                  % self.__class__)

    def getLeftFooter(self):
        "The bottom left text, SoL description and version by default."

        from pkg_resources import get_distribution

        dist = get_distribution('hurm.fe')
        description = dist.project_name
        version = dist.version

        return '%s %s %s' % (description, gettext('version'), version)

    def getRightFooter(self):
        "The bottom right text, current time by default."

        # TRANSLATORS: this is a Python strftime() format, see
        # http://docs.python.org/3/library/time.html#time.strftime
        return self.timestamp.strftime(str(gettext('%m-%d-%Y %I:%M %p')))

    def getCenterFooter(self):
        "The bottom center text, current page number by default."

        if self.doc.page > 1:
            return self.getSubTitle() + ', ' + gettext('page %d') % self.doc.page
        else:
            return gettext('page %d') % self.doc.page

    def decoratePage(self, canvas, doc):
        "Add standard decorations to the current page."

        canvas.saveState()
        canvas.setFont(BASE_FONT_NAME, 6)
        w, h = doc.pagesize
        hh = doc.bottomMargin + doc.height + doc.topMargin/2
        hl = doc.leftMargin
        hc = doc.leftMargin + doc.width/2.0
        hr = doc.leftMargin + doc.width
        canvas.drawString(hl, hh, self.getLeftHeader())
        canvas.drawCentredString(hc, hh, self.getCenterHeader())
        canvas.drawRightString(hr, hh, self.getRightHeader())
        fh = doc.bottomMargin/2
        canvas.drawString(hl, fh, self.getLeftFooter())
        canvas.drawCentredString(hc, fh, self.getCenterFooter())
        canvas.drawRightString(hr, fh, self.getRightFooter())
        canvas.restoreState()

    def execute(self, request):
        """Create and build the document.

        :param request: the Pyramid request instance
        """

        self.createDocument()
        self.doc.build(list(self.getElements()))

    def getElements(self):
        "Return a list or an iterator of all the elements."

        raise NotImplementedError("%s should implement this method!"
                                  % self.__class__)


class ParameterError(ValueError):
    pass


class BasicEditionPrintout(BasicPrintout):
    @classmethod
    def getArgumentsFromRequest(klass, session, request):
        args = super().getArgumentsFromRequest(session, request)
        t = args[0]
        try:
            idedition = int(request.matchdict['idedition'])
        except KeyError:
            raise ParameterError(t(_('Missing argument: $name',
                                     mapping=dict(name='idedition'))))

        edition = session.query(Edition).get(idedition)
        if edition is None:
            raise ParameterError(t(_('No edition with id $id',
                                     mapping=dict(id=str(idedition)))))

        args.append(edition)
        return args

    def __init__(self, output, translate, pluralize, edition):
        super().__init__(output, translate, pluralize)
        self.edition = edition

    def getLeftHeader(self):
        "Return edition's description."

        return self.edition.description

    def getRightHeader(self):
        "Return edition's period."

        # TRANSLATORS: this is a Python strftime() format, see
        # http://docs.python.org/3/library/time.html#time.strftime
        dateformat = gettext('%m-%d-%Y')

        return "%s — %s" % (self.edition.startdate.strftime(dateformat),
                            self.edition.enddate.strftime(dateformat))

    def getElements(self):
        "Yield basic elements for the title frame in the first page."

        title = self.getTitle()
        tstyle, ststyle = reduce_fontsize_to_fit_width(title, self.title_width - 1*cm,
                                                       title_style, subtitle_style)

        yield Paragraph(title, tstyle)
        yield Paragraph(self.getSubTitle(), ststyle)
        yield FrameBreak()
        yield NextPageTemplate('laterPages')


def create_pdf(session, request, maker):
    try:
        output = mktemp(prefix='hurm')
        args = maker.getArgumentsFromRequest(session, request)
        builder = maker(output, *args)
        try:
            builder.execute(request)

            f = open(output, 'rb')
            content = f.read()
            f.close()
        finally:
            try:
                unlink(output)
            except OSError:
                pass

        response = request.response
        response.content_type = 'application/pdf'
        cdisp = 'attachment; filename=%s.pdf' % maker.__name__
        response.content_disposition = cdisp
        response.body = content
        response.cache_control.public = True
        response.cache_control.max_age = builder.cache_max_age
        return response
    except ParameterError as e:
        logger.error("Couldn't create report %s: %s", maker.__name__, e)
        raise HTTPBadRequest(str(e))
    except Exception as e:
        logger.critical("Couldn't create report %s: %s", maker.__name__, e, exc_info=True)
        raise HTTPInternalServerError(str(e))


def render_timedelta(tdelta):
    """Convert a datetime.timedelta to a compact string.

    Examples::

      >>> import datetime
      >>> td = datetime.timedelta(hours=5, minutes=30)
      >>> print render_timedelta(td)
      5:30
      >>> td = datetime.timedelta(days=1, hours=5, minutes=45)
      >>> print render_timedelta(td)
      29:45
      >>> td = datetime.timedelta(days=-1, hours=4, minutes=15)
      >>> print render_timedelta(td)
      -19:45
      >>> print render_timedelta(None)
      0:00
    """

    if not tdelta:
        return '0:00'

    days = tdelta.days
    secs = tdelta.seconds

    if days<0:
        minus = "-"
        secs = 24*60*60 - secs
        days = -days - 1
    else:
        minus = ""

    secs += days * 24*60*60
    secs /= 60
    mm = secs % 60
    secs /= 60
    hh = secs

    return "%s%d:%02d" % (minus, hh, mm)



# -*- coding: utf-8 -*-
# :Project:   hurm -- Authentication related views
# :Created:   lun 01 feb 2016 20:55:40 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

import logging

from pyramid.events import NewRequest, subscriber
from pyramid.httpexceptions import HTTPUnauthorized
from pyramid.i18n import get_localizer
from pyramid.view import view_config

from sqlalchemy.orm import exc as saexc

from hurm.db.tables import password_manager
from hurm.db.entities import DBSession, Person

from ..i18n import translatable_string as _
from . import get_request_logger


logger = logging.getLogger(__name__)

NO_SUCH_USER = _('No such user!')
WRONG_PASSWORD = _('Wrong password')
MANDATORY_FIELD = _('Mandatory field')
MISSING_FIELDS = _('Missing fields')

ADMIN_ROLE = "HuRM Admin"
"Special role to identify HuRM admins, those having full access"


@subscriber(NewRequest)
def check_authorized_request(event,
                             authorized_paths={'/',
                                               '/auth/login',
                                               '/catalog',
                                               '/extjs-l10n',
                                               }):
    """Assert the request is authorized.

    This function gets hooked at the Pyramid's ``NewRequest`` event,
    so it will be executed at the start of each new request.

    If the user has been authenticated, or if she is requesting a
    static resource or one of the authentication views, then nothing
    happens. Otherwise an HTTPUnauthorized exception is raised.
    """

    request = event.request
    rpath = request.path
    sw = rpath.startswith

    # Authenticated user?
    session = request.session
    if 'user_id' in session:
        t = get_localizer(request).translate
        if sw('/data/persons') and not session['is_admin']:
            raise HTTPUnauthorized(t(_('Unauthorized access')))
        return

    # Anonymous authorized path or static resource?
    if rpath in authorized_paths or sw('/static/') or sw('/desktop/'):
        return

    if sw('/scripts') and request.registry.settings.get('desktop.debug', False):
        return

    t = get_localizer(request).translate
    raise HTTPUnauthorized(t(_('You must re-authenticate yourself')))


MODULES = (
    dict(classname='HuRM.module.Activities'),
    dict(classname='HuRM.module.Availabilities'),
    dict(classname='HuRM.module.Duties'),
    dict(classname='HuRM.module.Editions',
         shortcut=dict(
             name=_('Editions'),
             iconCls='editions-icon',
             moduleId='editions-win')),
    dict(classname='HuRM.module.LocationDuties'),
    dict(classname='HuRM.module.Locations',
         shortcut=dict(
             name=_('Locations'),
             iconCls='locations-icon',
             moduleId='locations-win')),
    dict(classname='HuRM.module.PersonDuties'),
    dict(classname='HuRM.module.Persons',
         shortcut=dict(
             name=_('Persons'),
             iconCls='persons-icon',
             moduleId='persons-win')),
    dict(classname='HuRM.module.Tasks'),
    )


@view_config(route_name='login', renderer='json')
def _auth_user(request):
    t = get_localizer(request).translate

    data = request.params

    email = data.get('email', None)
    password = data.get('password', None)

    if email and password:
        sasess = DBSession()
        try:
            user = sasess.query(Person).filter_by(email=email).one()
        except saexc.NoResultFound:
            user = None

        if user is None or not password_manager.check(user.password, password):
            # Do not give hints about either fields
            get_request_logger(request, logger).warning('Unsuccessful login attempt for %s',
                                                        email)
            return {'success': False,
                    'message': t(NO_SUCH_USER),
                    'errors': {'email': t(NO_SUCH_USER),
                               'password': t(WRONG_PASSWORD)}}


        s = request.session
        is_admin = s['is_admin'] = ADMIN_ROLE in user.role
        user_id = s['user_id'] = user.idperson
        s['user_name'] = user.email
        s['user_fn'] = user.firstname
        s['user_ln'] = user.lastname
        fullname = s['user_fullname'] = t(user.fullname)

        get_request_logger(request, logger).info('New %ssession for %s',
                                                 "admin " if is_admin else "", fullname)

        def translate_name(cfg):
            copy = dict(cfg)
            copy['name'] = t(copy['name'])
            return copy

        user_modules = [m for m in MODULES
                        if is_admin or m['classname'] != 'HuRM.module.Persons']

        return {'success': True,
                'fullname': fullname,
                'is_admin': is_admin,
                'user_id': user_id,
                'modules': [m['classname'] for m in user_modules],
                'shortcuts': [translate_name(m['shortcut'])
                              for m in user_modules
                              if 'shortcut' in m],
                'quickstart': [translate_name(m['quickstart'])
                               for m in user_modules
                               if 'quickstart' in m]}
    else:
        errors = {}
        if not email:
            errors['email'] = t(MANDATORY_FIELD)
        if not password:
            errors['password'] = t(MANDATORY_FIELD)
        return {'success': False,
                'message': t(MISSING_FIELDS),
                'errors': errors}


@view_config(route_name='logout', renderer='json')
def _logout(request):
    get_request_logger(request, logger).info('Session terminated')
    request.session.invalidate()
    return {'success': True, 'message': 'Goodbye'}



# -*- coding: utf-8 -*-
# :Project:   hurm -- Tests for the data views
# :Created:   dom 07 feb 2016 18:38:43 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

import pytest

from metapensiero.sqlalchemy.proxy.json import py2json

@pytest.mark.parametrize("view", [
    '/data/activities',
    '/data/availabilities',
    '/data/available_persons',
    '/data/countries',
    '/data/duties',
    '/data/editions',
    '/data/locations',
    '/data/persons',
    '/data/subcountries',
    '/data/tasks',
])
def test_fetch(app, view):
    response = app.get(view)
    result = response.json
    assert result['success'] is True
    assert result['message'] == "Ok"
    assert result['count'] > 0
    assert isinstance(result['root'], list)
    assert isinstance(result['root'][0], dict)


def test_available_persons_for_task(app, test_edition):
    task = test_edition.tasks[0]
    response = app.get('/data/available_persons?idtask=%d' % task.idtask)
    result = response.json
    assert result['success'] is True
    assert result['message'] == "Ok"
    assert result['count'] > 0
    assert isinstance(result['root'], list)
    assert isinstance(result['root'][0], dict)


@pytest.mark.parametrize("view", [
    '/data/activities',
    '/data/availabilities',
    '/data/duties',
    '/data/editions',
    '/data/locations',
    '/data/persons',
    '/data/tasks',
])
def test_update(app, view):
    response = app.get(view, {'limit': 1, 'metadata': 'metadata'})
    result = response.json
    assert result['success'] is True
    assert result['message'] == "Ok"
    assert isinstance(result['root'], list)
    assert len(result['root']) == 1

    pkname = result['metadata']['primary_key']
    pkvalue = result['root'][0][pkname]
    change = (pkname, {pkname: pkvalue, 'note': 'Foo bar'})
    response = app.post(view, {'modified_records': py2json([change]),
                               'deleted_records': py2json([])})
    result = response.json
    assert result['success'] is True
    assert result['message'] == "Ok"

    response = app.get(view, {'limit': 1, 'filter_' + pkname: pkvalue})
    result = response.json
    assert result['success'] is True
    assert result['message'] == "Ok"
    assert result['root'][0]['note'] == 'Foo bar'



# -*- coding: utf-8 -*-
# :Project:   hurm -- Test for the PDF generation views
# :Created:   ven 12 feb 2016 10:26:38 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from datetime import timedelta

import pytest

from hurm.fe.printouts import render_timedelta


def test_person_duties(app, test_edition, jane_tree):
    response = app.get('/pdf/duties/edition/%d/person/%d' %
                       (test_edition.idedition, jane_tree.idperson))
    assert response.content_type == 'application/pdf'


def test_location_duties(app, test_edition, reception):
    response = app.get('/pdf/duties/edition/%d/location/%d' %
                       (test_edition.idedition, reception.idlocation))
    assert response.content_type == 'application/pdf'


def test_by_location_duties(app, test_edition):
    response = app.get('/pdf/duties/edition/%d/locations' % test_edition.idedition)
    assert response.content_type == 'application/pdf'


def test_by_person_duties(app, test_edition):
    response = app.get('/pdf/duties/edition/%d/persons' % test_edition.idedition)
    assert response.content_type == 'application/pdf'


def test_persons_duties_summary(app, test_edition):
    response = app.get('/pdf/duties/edition/%d/summary' % test_edition.idedition)
    assert response.content_type == 'application/pdf'


@pytest.mark.parametrize('tdelta,expected', [
    (timedelta(hours=5, minutes=30), '5:30'),
    (timedelta(days=1, hours=5, minutes=45), '29:45'),
    (timedelta(days=-1, hours=4, minutes=15), '-19:45'),
    (None, '0:00'),
])
def test_render_timedelta(tdelta, expected):
    assert render_timedelta(tdelta) == expected



# -*- coding: utf-8 -*-
# :Project:   hurm -- Availabilities data view
# :Created:   mar 02 feb 2016 23:41:55 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from pyramid.view import view_config

from sqlalchemy import func, select, text

from hurm.db.tables import availabilities, duties, persons, tasks

from ..i18n import translatable_string as _
from . import expose


_a = availabilities.alias('a')
_d = duties.alias('d')
_p = persons.alias('p')
_t = tasks.alias('t')

_nduties = (
    select([func.count()],
           from_obj=_d.join(_t))
    .where(_t.c.idedition == _a.c.idedition)
    .where(_t.c.date == _a.c.date)
    .where(_d.c.idperson == _p.c.idperson)
    .where(_d.c.starttime >= func.coalesce(_a.c.starttime, '00:00'))
    .where(_d.c.endtime <= func.coalesce(_a.c.endtime, '24:00'))
)

_ptime = (
    select([func.to_char(func.sum(_d.c.endtime - _d.c.starttime), 'HH24:MI')],
           from_obj=_d.join(_t))
    .where(_t.c.idedition == _a.c.idedition)
    .where(_t.c.date == _a.c.date)
    .where(_d.c.idperson == _p.c.idperson)
    .where(_d.c.starttime >= func.coalesce(_a.c.starttime, '00:00'))
    .where(_d.c.endtime <= func.coalesce(_a.c.endtime, '24:00'))
)

@view_config(route_name='availabilities', renderer='json')
@expose(select([_a.c.idavailability,
                _a.c.idedition,
                _a.c.idperson,
                # text() is needed here, to workaround default normalization
                # performed by the Name type decorator, sigh...
                (_p.c.lastname + text("' '") + _p.c.firstname).label('Person'),
                _a.c.date,
                _a.c.starttime,
                _a.c.endtime,
                _a.c.note,
                _nduties.as_scalar().label('NDuties'),
                _ptime.as_scalar().label('Time')],
               from_obj=_a.join(_p)),
        metadata=dict(
            NDuties=dict(
                hint=_("Number of duties assigned to the person"),
                label=_("Duties"),
            ),
            Person=dict(
                flex=1,
                hint=_('Full name of the person'),
                label=_('Person'),
                lookup=dict(url='/data/available_persons',
                            idField='idperson',
                            displayField='fullname'),
            ),
            Time=dict(
                label=_('Time'),
                hint=_('Assigned time'),
                width=50,
            ),
            endtime=dict(min='07:00'),
            starttime=dict(min='07:00'),
            note=dict(width=120),
        ))
def _availabilities(request, results):
    return results



# -*- coding: utf-8 -*-
# :Project:   hurm -- Locations data view
# :Created:   mar 02 feb 2016 18:50:01 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from pyramid.view import view_config

from sqlalchemy import func, select

from hurm.db.tables import locations, tasks

from ..i18n import translatable_string as _, translator
from . import expose


_l = locations.alias('l')
_t = tasks.alias('t')

@view_config(route_name='locations', renderer='json')
@expose(select([_l.c.idlocation,
                _l.c.description,
                _l.c.address,
                _l.c.city,
                _l.c.province,
                _l.c.zip,
                _l.c.country,
                _l.c.latitude,
                _l.c.longitude,
                _l.c.phone,
                _l.c.mobile,
                _l.c.email,
                _l.c.note,
                select([func.count(_t.c.idtask)],
                       _t.c.idlocation == _l.c.idlocation)
                .as_scalar().label('Tasks')]),
        metadata=dict(
            Tasks=dict(
                hidden=True,
                label=_('Tasks'),
                hint=_('Number of tasks associated to the location, in any edition.')
            ),
            address=dict(width=150),
            city=dict(width=70),
            country=dict(hidden=True),
            description=dict(flex=1),
            email=dict(hidden=True, width=100, vtype='email'),
            latitude=dict(hidden=True),
            longitude=dict(hidden=True),
            mobile=dict(hidden=True, vtype='phone'),
            note=dict(width=120),
            phone=dict(hidden=True, vtype='phone'),
            province=dict(hidden=True),
            zip=dict(hidden=True),
        ))
def _locations(request, results):
    from gettext import translation
    from pycountry import LOCALES_DIR, countries, subdivisions

    if 'metadata' in results:
        t = translator(request)
        results['metadata']['fields'].append({
            'label': t(_('Country')),
            'hint': t(_('The name of the country')),
            'name': 'Country',
            'nullable': True,
            'hidden': True,
            'lookup': dict(url='/data/countries',
                           remoteFilter=False,
                           lookupField='country',
                           idField='code',
                           displayField='name'),
        })
        results['metadata']['fields'].append({
            'label': t(_('Province')),
            'hint': t(_('The name of the province')),
            'name': 'Province',
            'nullable': True,
            'hidden': True,
            'lookup': dict(url='/data/subcountries',
                           remoteFilter=False,
                           lookupField='province',
                           idField='code',
                           displayField='name'),
        })

    if 'root' in results:
        lname = request.locale_name
        try:
            tc = translation('iso3166', LOCALES_DIR, languages=[lname]).gettext
        except IOError:
            tc = lambda x: x
        try:
            tp = translation('iso3166_2', LOCALES_DIR, languages=[lname]).gettext
        except IOError:
            tp = lambda x: x

        locations = results['root']
        for location in locations:
            code = location.get('country', False)
            if code:
                location['Country'] = tc(countries.get(alpha2=code).name)
            elif code is not False:
                location['Country'] = None
            code = location.get('province', False)
            if code:
                location['Province'] = tp(subdivisions.get(code=code).name)
            elif code is not False:
                location['Province'] = None

    return results



# -*- coding: utf-8 -*-
# :Project:   hurm -- Download related views
# :Created:   mer 17 feb 2016 09:40:27 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

import datetime
import logging

from pyramid.httpexceptions import HTTPBadRequest, HTTPInternalServerError
from pyramid.view import view_config

import yaml

from hurm.db.entities import Edition

from .. import DBSession
from ..i18n import translatable_string as _, translator


logger = logging.getLogger(__name__)


def _dump_edition(edition):
    from hurm.db.utils import dump

    return dump(edition, [{
        'entity': 'hurm.db.entities.edition.Edition',
        'key': 'description',
        'other': 'startdate,enddate,tasks'.split(',')
    }, {
        'entity': 'hurm.db.entities.activity.Activity',
        'key': 'description'
    }, {
        'entity': 'hurm.db.entities.location.Location',
        'key': 'description',
        'other': 'address,city,country,province,zip'.split(',')
    }, {
        'entity': 'hurm.db.entities.task.Task',
        'key': 'edition,activity,location,date,starttime'.split(','),
        'other': 'endtime,npersons,duties'.split(',')
    }, {
        'entity': 'hurm.db.entities.person.Person',
        'key': 'lastname,firstname'.split(','),
        'other': 'birthdate,mobile,phone,email,role,availabilities,preferred_activities'.split(',')
    }, {
        'entity': 'hurm.db.entities.personactivity.PersonActivity',
        'key': 'person,activity'.split(','),
    }, {
        'entity': 'hurm.db.entities.availability.Availability',
        'key': 'edition,person,date,starttime'.split(','),
        'other': 'endtime'
    }, {
        'entity': 'hurm.db.entities.duty.Duty',
        'key': 'person,task,starttime'.split(','),
        'other': 'endtime'
    }])


@view_config(route_name='yaml_dump')
def _yaml_dump(request):
    t = translator(request)

    try:
        idedition = int(request.matchdict['idedition'])
    except KeyError:
        error = t(_('Missing argument: $name', mapping=dict(name='idedition')))
        logger.error("Couldn't dump: %s", error)
        raise HTTPBadRequest(str(error))

    session = DBSession()

    edition = session.query(Edition).get(idedition)
    if edition is None:
        error = t(_('No edition with id $id', mapping=dict(id=str(idedition))))
        logger.error("Couldn't dump: %s", error)
        raise HTTPBadRequest(str(error))

    try:
        data = _dump_edition(edition)

        yaml.add_representer(datetime.time,
                             lambda dumper, data: dumper.represent_scalar('!time', str(data)))

        asyaml = yaml.dump(data, default_flow_style=False)

        response = request.response
        response.content_type = 'application/x-yaml'
        cdisp = 'attachment; filename=edition-%s.yaml' % edition.idedition
        response.content_disposition = cdisp
        response.text = asyaml
        return response
    except Exception as e:
        logger.critical("Couldn't dump: %s", e, exc_info=True)
        raise HTTPInternalServerError(str(e))


def _create_workbook(translate, data):
    from io import BytesIO
    import xlwt

    byclass = {e['entity'].split('.')[-1]: e for e in data}

    output = BytesIO()

    workbook = xlwt.Workbook()
    heading_xf = xlwt.easyxf('font: bold on; align: wrap on, vert centre, horiz center')
    date_xf = xlwt.easyxf(num_format_str='D-MMM-YY')
    time_xf = xlwt.easyxf(num_format_str='HH:MM')

    editions_sn = translate(_('Editions'))
    sheet = workbook.add_sheet(editions_sn)
    headings = (
        translate(_('Edition')),
        translate(_('Start date')),
        translate(_('End date')),
    )
    sheet.set_panes_frozen(True)
    sheet.set_horz_split_pos(1)
    sheet.set_remove_splits(True)

    for col, value in enumerate(headings):
        sheet.write(0, col, value, heading_xf)

    for row, edata in enumerate(byclass['Edition']['data'], 1):
        edata[':row'] = row+1
        sheet.write(row, 0, edata['description'])
        sheet.write(row, 1, edata['startdate'], date_xf)
        sheet.write(row, 2, edata['enddate'], date_xf)

    activities_sn = translate(_('Activities'))
    sheet = workbook.add_sheet(activities_sn)
    headings = (translate(_('Activity')),)
    sheet.set_panes_frozen(True)
    sheet.set_horz_split_pos(1)
    sheet.set_remove_splits(True)

    for col, value in enumerate(headings):
        sheet.write(0, col, value, heading_xf)

    for row, edata in enumerate(byclass['Activity']['data'], 1):
        edata[':row'] = row+1
        sheet.write(row, 0, edata['description'])

    locations_sn = translate(_('Locations'))
    sheet = workbook.add_sheet(locations_sn)
    headings = (
        translate(_('Location')),
        translate(_('Address')),
        translate(_('Zip')),
        translate(_('City')),
    )
    sheet.set_panes_frozen(True)
    sheet.set_horz_split_pos(1)
    sheet.set_remove_splits(True)

    for col, value in enumerate(headings):
        sheet.write(0, col, value, heading_xf)

    for row, edata in enumerate(byclass['Location']['data'], 1):
        edata[':row'] = row+1
        sheet.write(row, 0, edata['description'])
        sheet.write(row, 1, edata['address'])
        sheet.write(row, 2, edata['zip'])
        sheet.write(row, 3, edata['city'])

    persons_sn = translate(_('Persons'))
    sheet = workbook.add_sheet(persons_sn)
    headings = (
        translate(_('Full name')),
        translate(_('Email')),
        translate(_('Phone')),
        translate(_('Mobile')),
        translate(_('Birthdate')),
        translate(_('Role')),
        translate(_('Preferred activities')),
    )
    sheet.set_panes_frozen(True)
    sheet.set_horz_split_pos(1)
    sheet.set_remove_splits(True)

    for col, value in enumerate(headings):
        sheet.write(0, col, value, heading_xf)

    pref_acts = {}
    for pa in byclass['PersonActivity']['data']:
        p = pa['person']
        name = p['lastname']+p['firstname']
        a = pa['activity']
        pref_acts.setdefault(name, []).append(a['description'])

    for row, edata in enumerate(byclass['Person']['data'], 1):
        edata[':row'] = row+1
        sheet.write(row, 0, edata['lastname'] + ' ' + edata['firstname'])
        if 'email' in edata: sheet.write(row, 1, edata['email'])
        if 'phone' in edata: sheet.write(row, 2, edata['phone'])
        if 'mobile' in edata: sheet.write(row, 3, edata['mobile'])
        if 'birthdate' in edata: sheet.write(row, 4, edata['birthdate'], date_xf)
        if 'role' in edata: sheet.write(row, 5, edata['role'])
        sheet.write(row, 6, ', '.join(sorted(pref_acts.get(edata['lastname']+edata['firstname'], []))))

    availabilities_sn = translate(_('Availabilities'))
    sheet = workbook.add_sheet(availabilities_sn)
    headings = (
        translate(_('Person')),
        translate(_('Date')),
        translate(_('Start time')),
        translate(_('End time')),
    )
    sheet.set_panes_frozen(True)
    sheet.set_horz_split_pos(1)
    sheet.set_remove_splits(True)

    for col, value in enumerate(headings):
        sheet.write(0, col, value, heading_xf)

    for row, edata in enumerate(byclass['Availability']['data'], 1):
        edata[':row'] = row+1
        formula = "'%(sheet)s'!$A$%(row)d" % dict( sheet=persons_sn,
                                                   row=edata['person'][':row'])
        sheet.write(row, 0, xlwt.Formula(formula))
        sheet.write(row, 1, edata['date'], date_xf)
        if 'starttime' in edata: sheet.write(row, 2, edata['starttime'], time_xf)
        if 'endtime' in edata: sheet.write(row, 3, edata['endtime'], time_xf)

    tasks_sn = translate(_('Tasks'))
    sheet = workbook.add_sheet(tasks_sn)
    headings = (
        translate(_('Activity')),
        translate(_('Location')),
        translate(_('Date')),
        translate(_('Start time')),
        translate(_('End time')),
    )
    sheet.set_panes_frozen(True)
    sheet.set_horz_split_pos(1)
    sheet.set_remove_splits(True)

    for col, value in enumerate(headings):
        sheet.write(0, col, value, heading_xf)

    for row, edata in enumerate(byclass['Task']['data'], 1):
        edata[':row'] = row+1
        formula = "'%(sheet)s'!$A$%(row)d" % dict(sheet=activities_sn,
                                                  row=edata['activity'][':row'])
        sheet.write(row, 0, xlwt.Formula(formula))
        formula = "'%(sheet)s'!$A$%(row)d" % dict(sheet=locations_sn,
                                                  row=edata['location'][':row'])
        sheet.write(row, 1, xlwt.Formula(formula))
        sheet.write(row, 2, edata['date'], date_xf)
        if 'starttime' in edata: sheet.write(row, 3, edata['starttime'], time_xf)
        if 'endtime' in edata: sheet.write(row, 4, edata['endtime'], time_xf)

    duties_sn = translate(_('Duties'))
    sheet = workbook.add_sheet(duties_sn)
    headings = (
        translate(_('Person')),
        translate(_('Activity')),
        translate(_('Location')),
        translate(_('Date')),
        translate(_('Start time')),
        translate(_('End time')),
    )
    sheet.set_panes_frozen(True)
    sheet.set_horz_split_pos(1)
    sheet.set_remove_splits(True)

    for col, value in enumerate(headings):
        sheet.write(0, col, value, heading_xf)

    for row, edata in enumerate(byclass['Duty']['data'], 1):
        edata[':row'] = row+1
        formula = "'%(sheet)s'!$A$%(row)d" % dict(sheet=persons_sn,
                                                  row=edata['person'][':row'])
        sheet.write(row, 0, xlwt.Formula(formula))
        formula = "'%(sheet)s'!$A$%(row)d" % dict(sheet=activities_sn,
                                                  row=edata['task']['activity'][':row'])
        sheet.write(row, 1, xlwt.Formula(formula))
        formula = "'%(sheet)s'!$A$%(row)d" % dict(sheet=locations_sn,
                                                  row=edata['task']['location'][':row'])
        sheet.write(row, 2, xlwt.Formula(formula))
        sheet.write(row, 3, edata['task']['date'], date_xf)
        if 'starttime' in edata: sheet.write(row, 4, edata['starttime'], time_xf)
        if 'endtime' in edata: sheet.write(row, 5, edata['endtime'], time_xf)

    workbook.save(output)

    return output.getvalue()


@view_config(route_name='excel_dump')
def _excel_dump(request):
    t = translator(request)

    try:
        idedition = int(request.matchdict['idedition'])
    except KeyError:
        error = t(_('Missing argument: $name', mapping=dict(name='idedition')))
        logger.error("Couldn't dump: %s", error)
        raise HTTPBadRequest(str(error))

    session = DBSession()

    edition = session.query(Edition).get(idedition)
    if edition is None:
        error = t(_('No edition with id $id', mapping=dict(id=str(idedition))))
        logger.error("Couldn't dump: %s", error)
        raise HTTPBadRequest(str(error))

    try:
        data = _dump_edition(edition)
        response = request.response
        response.content_type = 'application/vnd.ms-excel'
        cdisp = 'attachment; filename=edition-%s.xls' % edition.idedition
        response.content_disposition = cdisp
        response.body = _create_workbook(t, data)
        return response
    except Exception as e:
        logger.critical("Couldn't dump: %s", e, exc_info=True)
        raise HTTPInternalServerError(str(e))



# -*- coding: utf-8 -*-
# :Project:   hurm -- Duties data view
# :Created:   mer 03 feb 2016 21:19:58 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from pyramid.view import view_config

from sqlalchemy import select, text

from hurm.db.tables import activities, duties, locations, persons, tasks

from ..i18n import translatable_string as _
from . import expose

_a = activities.alias('a')
_d = duties.alias('d')
_l = locations.alias('l')
_p = persons.alias('p')
_t = tasks.alias('t')

query = select(
    [_d.c.idduty,
     _d.c.idtask,
     _d.c.idperson,
     # text() is needed here, to workaround default normalization
     # performed by the Name type decorator, sigh...
     (_p.c.lastname + text("' '") + _p.c.firstname).label('Person'),
     _t.c.date,
     _d.c.starttime,
     _d.c.endtime,
     _l.c.description.label('Location'),
     _a.c.description.label('Activity'),
     _d.c.note],
    from_obj=_d.join(_p).join(_t).join(_a).join(_l))

@view_config(route_name='duties', renderer='json')
@expose(query,
        metadata=dict(
            Activity=dict(width=120),
            Location=dict(flex=1),
            Person=dict(
                flex=1,
                hint=_('Full name of the person'),
                label=_('Person'),
                lookup=dict(
                    displayField='fullname',
                    idField='idperson',
                    url='/data/available_persons',
                    otherFields='Preferred',
                    innerTpl='<div><tpl if="Preferred"><strong></tpl>{fullname}<tpl if="Preferred"></strong></tpl></div>',
                ),
            ),
            endtime=dict(min='07:00'),
            note=dict(width=120),
            starttime=dict(min='07:00'),
        ))
def _duties(request, results):
    return results



# -*- coding: utf-8 -*-
# :Project:   hurm -- Editions data view
# :Created:   mar 02 feb 2016 16:54:48 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from pyramid.view import view_config

from sqlalchemy import func, select

from hurm.db.tables import editions, tasks

from ..i18n import translatable_string as _
from . import expose


_e = editions.alias('e')
_t = tasks.alias('t')

@view_config(route_name='editions', renderer='json')
@expose(select([_e.c.idedition,
                _e.c.description,
                _e.c.startdate,
                _e.c.enddate,
                _e.c.note,
                select([func.count(_t.c.idtask)],
                       _t.c.idedition == _e.c.idedition)
                .as_scalar().label('Tasks')]),
        metadata=dict(
            Tasks=dict(
                label=_('Tasks'),
                hint=_('Number of tasks defined for the edition.')
            ),
            description=dict(flex=1),
            note=dict(width=120),
        ))
def _editions(request, results):
    return results



# -*- coding: utf-8 -*-
# :Project:   hurm -- Tasks data view
# :Created:   mer 03 feb 2016 19:02:42 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from pyramid.view import view_config

from sqlalchemy import select, Boolean, Integer
from sqlalchemy.sql.expression import literal_column

from hurm.db.tables import activities, locations, tasks

from ..i18n import translatable_string as _
from . import expose


_a = activities.alias('a')
_l = locations.alias('l')
_t = tasks.alias('t')

@view_config(route_name='tasks', renderer='json')
@expose(select([_t.c.idtask,
                _t.c.idedition,
                _t.c.idlocation,
                _l.c.description.label('Location'),
                _t.c.date,
                _t.c.starttime,
                _t.c.endtime,
                _t.c.idactivity,
                _a.c.description.label('Activity'),
                _t.c.npersons,
                literal_column(
                    "not exists (select 1 from task_timeline(t.idtask) tl"
                    " where tl.npersons < t.npersons)"
                    , Boolean).label('Complete'),
                literal_column(
                    "cast(100 *"
                    " extract(epoch from (select sum((tl.endtime-tl.starttime)*tl.npersons)"
                    " from task_timeline(t.idtask) tl))"
                    " /"
                    " extract(epoch from (coalesce(t.endtime, '24:00')-t.starttime)*t.npersons) as integer)"
                    , Integer).label('Coverage'),
                _t.c.note],
               from_obj=_t.join(_l).join(_a)),
        metadata=dict(
            Activity=dict(
                hint=_('Description of the activity'),
                label=_('Activity'),
                lookup=dict(url='/data/activities?only_cols=idactivity,description'
                            '&sort=description',
                            idField='idactivity',
                            displayField='description'),
                width=100,
            ),
            Complete=dict(
                hint=_('Whether the task is completely covered'),
                label=_('Complete'),
                width=60,
            ),
            Coverage=dict(
                hint=_('Percentage of time covered by assigned duties'),
                label=_('%'),
                width=50,
            ),
            Location=dict(
                flex=1,
                hint=_('Description of the location'),
                label=_('Location'),
                lookup=dict(url='/data/locations?only_cols=idlocation,description'
                            '&sort=description',
                            idField='idlocation',
                            displayField='description'),
            ),
            endtime=dict(min='07:00'),
            note=dict(width=120),
            npersons=dict(min=1),
            starttime=dict(min='07:00'),
        ))
def _tasks(request, results):
    return results



# -*- coding: utf-8 -*-
# :Project:   hurm -- Edition related printouts
# :Created:   dom 14 feb 2016 19:27:54 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from datetime import datetime, timedelta
from itertools import groupby

from pyramid.view import view_config

from reportlab.lib import colors
from reportlab.lib.pagesizes import A4, landscape
from reportlab.lib.units import cm
from reportlab.platypus import CondPageBreak, KeepTogether, Paragraph, Spacer

from reportlab.platypus import TableStyle
from reportlab.platypus.tables import Table

from sqlalchemy.orm import object_session

from hurm.db.entities import Location, Task

from .. import DBSession
from ..i18n import translatable_string as _
from . import BasicEditionPrintout, create_pdf, render_timedelta, subtitle_style


class ByLocationDuties(BasicEditionPrintout):
    pagesize = landscape(A4)

    def __init__(self, output, translate, pluralize, edition):
        super().__init__(output, translate, pluralize, edition)
        session = object_session(edition)
        self.locations = session.query(Location).join(Task) \
                         .filter(Task.idedition == edition.idedition) \
                         .order_by(Location.description) \
                         .all()

    def getCenterHeader(self):
        return self.getTitle()

    def getTitle(self):
        return self.translate(_('Duties calendar by location'))

    def getSubTitle(self):
        ngettext = self.pluralize
        count = len(self.locations)
        return ngettext('$count location', '$count locations', count,
                        mapping=dict(count=count))

    def getElements(self):
        from .location import duties_table

        yield from super().getElements()

        for location in self.locations:
            yield Spacer(0, 10)
            yield KeepTogether([Paragraph(location.description, subtitle_style),
                                Spacer(0, 5),
                                duties_table(self.translate, self.doc, location, self.edition)])


@view_config(route_name='pdf_by_location_duties')
def _by_location_duties(request):
    return create_pdf(DBSession(), request, ByLocationDuties)


class ByPersonDuties(BasicEditionPrintout):
    pagesize = landscape(A4)

    def __init__(self, output, translate, pluralize, edition):
        super().__init__(output, translate, pluralize, edition)
        persons = set(duty.person for task in edition.tasks for duty in task.duties)
        self.persons = sorted(persons, key=lambda p: self.translate(p.fullname))

    def getCenterHeader(self):
        return self.getTitle()

    def getTitle(self):
        return self.translate(_('Duties calendar by person'))

    def getSubTitle(self):
        ngettext = self.pluralize
        count = len(self.persons)
        return ngettext('$count person', '$count persons', count,
                        mapping=dict(count=count))

    def getElements(self):
        from .person import duties_table

        yield from super().getElements()

        for person in self.persons:
            yield Spacer(0, 10)
            yield KeepTogether([Paragraph(self.translate(person.fullname), subtitle_style),
                                Spacer(0, 5),
                                duties_table(self.translate, self.doc, person)])


@view_config(route_name='pdf_by_person_duties')
def _by_person_duties(request):
    return create_pdf(DBSession(), request, ByPersonDuties)


class PersonsDutiesSummary(BasicEditionPrintout):
    def __init__(self, output, translate, pluralize, edition):
        from ..views.duties import query

        super().__init__(output, translate, pluralize, edition)
        conn = object_session(edition).connection()
        self.persons = conn.execute(query.order_by("Person", "date", "starttime")).fetchall()

    def getCenterHeader(self):
        return self.getTitle()

    def getTitle(self):
        return self.translate(_('Persons duties summary'))

    def getSubTitle(self):
        ngettext = self.pluralize
        count = len(set(p.Person for p in self.persons))
        return ngettext('$count person', '$count persons', count,
                        mapping=dict(count=count))

    def getElements(self):
        yield from super().getElements()

        # TRANSLATORS: this is a Python strftime() format, see
        # http://docs.python.org/3/library/time.html#time.strftime
        datefmt = self.translate(_('%m-%d-%Y'))

        # TRANSLATORS: this is the Python strftime() template used to format times,
        # see http://docs.python.org/3/library/time.html#time.strftime
        timefmt = self.translate(_('%I:%M %p'))

        combine = datetime.combine

        for person, dates in groupby(self.persons, lambda p: p.Person):
            styles = [
                ('ALIGN', (0, 0), (-1, 0), 'CENTER'),
                ('ALIGN', (0, 0), (1, -1), 'CENTER'),
                ('SIZE', (0, 0), (-1, 0), 14),
            ]

            rows = [
                [ self.translate(s) for s in (_('Date'), _('Start'), _('End'),
                                              _('Location'), _('Activity')) ],
            ]

            pdate = None
            dtime = timedelta(0)
            rownum = 1

            for d in dates:
                if pdate is not None and pdate != d.date:
                    rows.append([self.translate(_('Total time')), render_timedelta(dtime),
                                 '', '', ''])
                    styles.append(('LINEABOVE', (1, rownum), (2, rownum), 0.25, colors.black))
                    styles.append(('SPAN', (1, rownum), (2, rownum)))
                    rows.append([''] * len(rows[0]))
                    dtime = timedelta(0)
                    rownum += 2

                rows.append([d.date.strftime(datefmt) if pdate != d.date else '"',
                             d.starttime.strftime(timefmt),
                             d.endtime.strftime(timefmt),
                             d.Location,
                             d.Activity])
                rownum += 1

                dtime += (combine(d.date, d.endtime) - combine(d.date, d.starttime))
                pdate = d.date

            rows.append(['', render_timedelta(dtime), '', '', ''])
            styles.append(('LINEABOVE', (1, rownum), (2, rownum), 0.25, colors.black))
            styles.append(('SPAN', (1, rownum), (2, rownum)))
            rows.append([''] * len(rows[0]))

            yield CondPageBreak(4*cm)
            yield Spacer(0, 15)
            yield Paragraph(person, subtitle_style)
            yield Spacer(0, 5)
            yield Table(rows, style=TableStyle(styles), repeatRows=1)


@view_config(route_name='pdf_persons_duties_summary')
def _persons_duties_summary(request):
    return create_pdf(DBSession(), request, PersonsDutiesSummary)



# -*- coding: utf-8 -*-
# :Project:   hurm -- Activities data view
# :Created:   mar 02 feb 2016 17:36:07 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from pyramid.view import view_config

from sqlalchemy import func, select

from hurm.db.tables import activities, tasks

from ..i18n import translatable_string as _
from . import expose


_a = activities.alias('a')
_t = tasks.alias('t')

@view_config(route_name='activities', renderer='json')
@expose(select([_a.c.idactivity,
                _a.c.description,
                _a.c.note,
                select([func.count(_t.c.idtask)],
                       _t.c.idactivity == _a.c.idactivity)
                .as_scalar().label('Tasks')]),
        metadata=dict(
            Tasks=dict(
                hidden=True,
                label=_('Tasks'),
                hint=_('Number of tasks associated to the activity, in any edition.')
            ),
            description=dict(flex=1),
            note=dict(width=120),
        ))
def _activities(request, results):
    return results



# -*- coding: utf-8 -*-
# :Project:   hurm -- Test for the downloads views
# :Created:   mer 17 feb 2016 13:23:58 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

def test_yaml_dump(app, test_edition):
    response = app.get('/yaml/edition/%d' % test_edition.idedition)
    assert response.content_type == 'application/x-yaml'


def test_excel_dump(app, test_edition):
    response = app.get('/xls/edition/%d' % test_edition.idedition)
    assert response.content_type == 'application/vnd.ms-excel'



# -*- coding: utf-8 -*-
# :Project:   hurm -- Package setup
# :Created:   lun 01 feb 2016 20:08:25 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

import os
from codecs import open

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    README = f.read()
with open(os.path.join(here, 'CHANGES.rst'), encoding='utf-8') as f:
    CHANGES = f.read()
with open(os.path.join(here, 'version.txt'), encoding='utf-8') as f:
    VERSION = f.read().strip()

requires = [
    'babel',
    'cryptacular',
    'metapensiero.extjs.desktop',
    'metapensiero.sqlalchemy.proxy',
    'pyramid',
    'pyramid_tm',
    'pyyaml',
    'reportlab',
    'setuptools',
    'sqlalchemy',
    'transaction',
    'waitress',
    'xlwt',
    'zope.sqlalchemy',
    ]

setup(
    name='hurm.fe',
    version=VERSION,
    description='Human Resources Manager, ExtJS frontend',
    long_description=README + '\n\n' + CHANGES,

    license="GPLv3+",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Framework :: Pyramid",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Programming Language :: JavaScript",
        "Operating System :: OS Independent",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved ::"
        " GNU General Public License v3 or later (GPLv3+)",
    ],
    author='Lele Gaifax',
    author_email='lele@metapensiero.it',
    url='https://bitbucket.org/lele/hurm.fe',
    keywords='web wsgi bfg pylons pyramid',

    packages=['hurm.fe'],
    namespace_packages=['hurm'],

    include_package_data=True,
    zip_safe=False,

    install_requires=requires,
    extras_require={'dev': ['metapensiero.tool.bump_version', 'readme']},

    entry_points="""\
    [paste.app_factory]
    main = hurm.fe:main
    """,
)



# -*- coding: utf-8 -*-
# :Project:   hurm -- Utility functions related to calendar printouts
# :Created:   sab 13 feb 2016 19:36:10 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from collections import namedtuple
from datetime import date, datetime, time, timedelta

from reportlab.lib import colors

from ..i18n import translatable_string as _


START_TIME = time(8, 0)
END_TIME = time(23, 59)
MINUTES_PER_CELL = 5
RESOLUTION = timedelta(minutes=MINUTES_PER_CELL)
COLS_PER_HOUR = 60 // MINUTES_PER_CELL


def calendar_header(hourfmt, font_size=6, min_time=START_TIME, max_time=END_TIME):
    """Generate the table's header row and the basic styles for a calendar printout."""

    dtcurr = datetime.combine(date.today(), min_time)
    dtstop = datetime.combine(date.today(), max_time)
    cells = [_('Date')]
    styles = [
        ('ALIGN', (0, 0), (-1, -1), 'CENTER'),
        ('VALIGN', (0, 0), (-1, -1), 'MIDDLE'),
        ('SIZE', (0, 1), (0, -1), 8),
        ('SIZE', (1, 1), (-1, -1), font_size),
        ('TOPPADDING', (1, 1), (-1, -1), 2),
        ('RIGHTPADDING', (1, 1), (-1, -1), 2),
        ('BOTTOMPADDING', (1, 1), (-1, -1), 2),
        ('LEFTPADDING', (1, 1), (-1, -1), 2),
    ]
    cellnumber = 1

    while dtcurr < dtstop:
        currtime = dtcurr.time()
        if currtime.minute == 0:
            cells.append(currtime.strftime(hourfmt))
            styles.append(('SPAN', (cellnumber, 0), (cellnumber+COLS_PER_HOUR-1, 0)))
        else:
            cells.append('')
        cellnumber += 1
        dtcurr += RESOLUTION

    return cells, styles


def _generate_cells(rownumber, firstcellnumber, content, start, stop,
                    min_time=START_TIME, max_time=END_TIME):
    "Generate row cells and styles in the given time span."

    dtcurr = datetime.combine(date.today(), start or min_time)
    dtstop = datetime.combine(date.today(), stop or max_time)
    cells = []
    ncells = 0
    styles = []

    if content:
        currcontent = '\n'.join(content)
    else:
        currcontent = ''

    while dtcurr < dtstop:
        cells.append(currcontent)
        if currcontent:
            currcontent = ''
        ncells += 1
        dtcurr += RESOLUTION

    if content and ncells > 1:
        styles.append(('BOX',
                       (firstcellnumber, rownumber),
                       (firstcellnumber+ncells-1, rownumber), 1, colors.black))
        styles.append(('SPAN',
                       (firstcellnumber, rownumber),
                       (firstcellnumber+ncells-1, rownumber)))

    return cells, styles, dtstop.time()


CellContent = namedtuple('CellContent', 'starttime, endtime, content')


def calendar_row(rownumber, cells, min_time=START_TIME, max_time=END_TIME):
    """Generate the table's row and styles for duties of a particular day."""

    lt = None
    rowcells = []
    rowstyles = []
    cellnumber = 1

    for cell in cells:
        st = cell.starttime
        et = cell.endtime

        cells, styles, lt = _generate_cells(rownumber, cellnumber, None,
                                            lt, st, min_time, max_time)
        rowcells.extend(cells)
        rowstyles.extend(styles)
        cellnumber += len(cells)

        cells, styles, lt = _generate_cells(rownumber, cellnumber, cell.content,
                                            lt, et, min_time, max_time)
        rowcells.extend(cells)
        rowstyles.extend(styles)
        cellnumber += len(cells)

    cells, styles, _ = _generate_cells(rownumber, cellnumber, None, lt, None,
                                       min_time, max_time)
    rowcells.extend(cells)
    rowstyles.extend(styles)

    return rowcells, rowstyles



# -*- coding: utf-8 -*-
# :Project:   hurm -- Person related printouts
# :Created:   gio 11 feb 2016 17:33:12 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

from itertools import groupby

from pyramid.view import view_config

from reportlab.lib.pagesizes import A4, landscape
from reportlab.platypus import TableStyle
from reportlab.platypus.tables import Table

from hurm.db.entities import Person

from .. import DBSession
from ..i18n import translatable_string as _
from . import BasicEditionPrintout, ParameterError, create_pdf
from .calendar import CellContent, calendar_header, calendar_row


def duties_table(translate, document, person):
    # TRANSLATORS: this is a Python strftime() format, see
    # http://docs.python.org/3/library/time.html#time.strftime
    datefmt = translate(_('%m-%d-%Y'))

    # TRANSLATORS: this is the Python strftime() template used to format times,
    # see http://docs.python.org/3/library/time.html#time.strftime
    timefmt = translate(_('%I:%M %p'))

    # TRANSLATORS: this is the strftime() template used to format hours,
    # see http://docs.python.org/3/library/time.html#time.strftime
    hourfmt = translate(_('%I %p'))

    min_time = min(duty.starttime for duty in person.duties)
    if min_time.minute != 0:
        min_time = min_time.replace(minute=0)
    max_time = max(duty.endtime for duty in person.duties)
    if max_time.minute != 0:
        max_time = max_time.replace(hour=max_time.hour+1, minute=0)

    cells, styles = calendar_header(hourfmt, min_time=min_time, max_time=max_time)
    rows = [cells]
    rownumber = 1

    for date, duties in groupby(person.duties, lambda d: d.task.date):
        ccells = [CellContent(duty.starttime, duty.endtime,
                              [duty.task.activity.description,
                               duty.task.location.description])
                 for duty in duties]
        for cell in ccells:
            if (cell.endtime.hour - cell.starttime.hour) > 1:
                cell.content.insert(0,
                                    cell.starttime.strftime(timefmt)
                                    + " — "
                                    + cell.endtime.strftime(timefmt))
            else:
                cell.content.insert(0, cell.starttime.strftime(timefmt))
                cell.content.insert(1, cell.endtime.strftime(timefmt))
        cells, rowstyles = calendar_row(rownumber, ccells,
                                        min_time=min_time, max_time=max_time)
        rows.append([date.strftime(datefmt)] + cells)
        styles.extend(rowstyles)
        rownumber += 1

    ntimecells = len(rows[0]) - 1
    cwidths = [None] + [(document.width - 40) // ntimecells] * ntimecells
    return Table(rows, cwidths, style=TableStyle(styles), repeatRows=1)


class PersonDuties(BasicEditionPrintout):
    pagesize = landscape(A4)

    @classmethod
    def getArgumentsFromRequest(klass, session, request):
        args = super().getArgumentsFromRequest(session, request)

        t = args[0]

        try:
            idperson = int(request.matchdict['idperson'])
        except KeyError:
            raise ParameterError(t(_('Missing argument: $name',
                                     mapping=dict(name='idperson'))))
        person = session.query(Person).get(idperson)
        if person is None:
            raise ParameterError(t(_('No person with id $id',
                                 mapping=dict(id=str(idperson)))))

        args.append(person)

        return args

    def __init__(self, output, translate, pluralize, edition, person):
        super().__init__(output, translate, pluralize, edition)
        self.person = person

    def getCenterHeader(self):
        fullname = self.translate(self.person.fullname)
        return self.translate(_('Duties calendar for $person',
                                mapping=dict(person=fullname)))

    def getTitle(self):
        return self.translate(self.person.fullname)

    def getSubTitle(self):
        count = len(self.person.duties)
        ngettext = self.pluralize
        return ngettext('$count duty', '$count duties', count,
                        mapping=dict(count=count))

    def getElements(self):
        yield from super().getElements()
        yield duties_table(self.translate, self.doc, self.person)


@view_config(route_name='pdf_person_duties')
def _person_duties(request):
    return create_pdf(DBSession(), request, PersonDuties)
